/***************************************************************************
                          knowit.cpp  -  description
                             -------------------
    begin                : czw wrz 26 08:27:40 CEST 2002
    copyright            : (C) 2002 by Micha Rudolf
    email                : mrudolf@mercury.ci.uw.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "knowit.h"
#include <qdatetime.h>
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qwhatsthis.h>
#include <qsimplerichtext.h>
#include <kcmdlineargs.h>
#include <kcolordialog.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <klineeditdlg.h>
#include <kprinter.h>
#include <kkeydialog.h>
#include <ktip.h>

const QString Knowit::Untitled = i18n("untitled") + ".kno";

Knowit::Knowit(QWidget*, const char *name) : KMainWindow(0, name),
   FindDlg(), PrefDlg(), ChooserDlg(), Notes()
{
   statusBar()->insertItem(i18n("Ready"), 0);
   Layout = new QSplitter(this);
   Items = new KListView(Layout, "Items");
   Edit = new QTextEdit(Layout, "Edit");
      
   connect(&FindDlg, SIGNAL(search()), this, SLOT(slotEditFind()));
   lastSought = 0;
   
   /* main view */
   setAutoSaveSettings();
   setCentralWidget(Layout);
      
   /* Items (KListView */
   Items->setMinimumSize(150, 150);
   Items->addColumn(i18n("Notes"));
   Items->setRootIsDecorated(true);
   Items->setItemsRenameable(true);
   Items->setDragEnabled(true);
   Items->setAcceptDrops(true);
   Items->setFullWidth();
   Items->setSorting(-1);
   /*
   KAction* a = new KAction(i18n("Add note from list"), "", "Ctrl+Return", this,
      SLOT(slotNoteAdd()), actionCollection(), "list_note_add");
   a->plug(Items);
   a = new KAction(i18n("Add subnote in notes list"), "", "Insert", this,
      SLOT(slotNoteAddChild()), actionCollection(), "list_note_add_child");
   a->plug(Items);
   a = new KAction(i18n("Delete subnote from notes list"), "", "Delete", this,
      SLOT(slotNoteRemove()), actionCollection(), "list_note_remove");
   a->plug(Items);
   */
   
   prevItem = 0;
   QWhatsThis::add(Items, i18n("<h2>Notes hierarchy</h2>"
      "Here you can browse notes tree. "
      "You can also add notes using right-click menu, or reorganize them using "
      "drag and drop."));
   QWhatsThis::add(Layout, i18n("Move it to change tree/edit layout."));
   connect(Items, SIGNAL(contextMenu(KListView*, QListViewItem*, const QPoint&)),
     this, SLOT(slotContextMenu(KListView*, QListViewItem*, const QPoint&)));
   connect(Items, SIGNAL(selectionChanged(QListViewItem*)), this,
     SLOT(slotItemChanged(QListViewItem*)));
   connect(Items, SIGNAL(selectionChanged(QListViewItem*)), this,
     SLOT(slotItemChanged(QListViewItem*)));

   /* Edit (QTextEdit) */
   Edit->setMinimumSize(200, 150);
   Edit->setTextFormat(Qt::RichText);
   QWhatsThis::add(Edit, i18n("<h2>Note text</h2>Add text for selected note here.\n"
      "Text can be formatted, for example <b>bold</b> or <i>italic</i>."));
   connect(Edit, SIGNAL(selectionChanged()), this, SLOT(slotEditChanged()));
   connect(Edit, SIGNAL(textChanged()), this, SLOT(slotEditChanged()));
   connect(Edit, SIGNAL(cursorPositionChanged(int, int)), this,
     SLOT(slotEditCursorChanged(int, int)));

   actionNew = KStdAction::openNew(this, SLOT(slotFileNew()), actionCollection());
   actionOpen = KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
   actionRecent = KStdAction::openRecent(this, SLOT(slotFileRecent(const KURL&)), actionCollection());
   actionSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
   actionSaveAs = KStdAction::saveAs(this, SLOT(slotFileSaveAs()), actionCollection());
   actionFileExport = new KAction(i18n("&Export to HTML..."), "", "", this,
      SLOT(slotFileExport()), actionCollection(), "file_export_html");
   actionPrint = KStdAction::print(this, SLOT(slotFilePrint()), actionCollection());
   actionFileInfo = new KAction(i18n("Document &information..."), "", "", this,
      SLOT(slotFileInfo()), actionCollection(), "file_info");
   actionQuit = KStdAction::quit(this, SLOT(slotFileQuit()),
      actionCollection());
      
   actionCut = KStdAction::cut(Edit, SLOT(cut()), actionCollection());
   connect(Edit, SIGNAL(copyAvailable(bool)), actionCut, SLOT(setEnabled(bool)));
   actionCut->setEnabled(false);
   
   actionCopy = KStdAction::copy(Edit, SLOT(copy()), actionCollection());
   connect(Edit, SIGNAL(copyAvailable(bool)), actionCopy, SLOT(setEnabled(bool)));
   actionCopy->setEnabled(false);
   
   actionPaste = KStdAction::paste(Edit, SLOT(paste()), actionCollection());
   
   actionUndo = KStdAction::undo(Edit, SLOT(undo()), actionCollection());
   connect(Edit, SIGNAL(undoAvailable(bool)), actionUndo, SLOT(setEnabled(bool)));

   actionRedo = KStdAction::redo(Edit, SLOT(redo()), actionCollection());
   connect(Edit, SIGNAL(redoAvailable(bool)), actionRedo, SLOT(setEnabled(bool)));
      
   actionBold = new KToggleAction(i18n("Toggle &bold"), "text_bold",
      "Ctrl+B", this, SLOT(slotEditBold()), actionCollection(), "edit_bold");
   actionItalic = new KToggleAction(i18n("Toggle &italic"), "text_italic",
      "Ctrl+I", this, SLOT(slotEditItalic()), actionCollection(),
      "edit_italic");
   actionUnderline = new KToggleAction(i18n("Toggle &underline"), "text_under",
      "Ctrl+U", this, SLOT(slotEditUnderline()), actionCollection(),
      "edit_underline");
   actionColor = new KAction(i18n("&Text color..."), "color_line", "Ctrl+#", this,
      SLOT(slotEditColor()), actionCollection(), "edit_color");
   actionSuperscript = new KAction(i18n("&Superscript"), "text_super", "", this,
      SLOT(slotEditSuperscript()), actionCollection(), "edit_superscript");
   actionSubscript = new KAction(i18n("&Subscript"), "text_sub", "", this,
      SLOT(slotEditSubscript()), actionCollection(), "edit_subscript");
   actionVNormal = new KAction(i18n("&Normal text"), "", "", this,
      SLOT(slotEditNormal()), actionCollection(), "edit_align_normal");
   actionFind = KStdAction::find(&FindDlg, SLOT(exec()), actionCollection());
   actionFindNext = KStdAction::findNext(this, SLOT(slotEditFindNext()), actionCollection());
   actionGoTo = new KAction(i18n("&Go to &note..."), "goto", "Ctrl+G", this,
      SLOT(slotEditGoTo()), actionCollection(), "goto_note");
   actionEditAlignLeft = new KToggleAction(i18n("Align &left"), "text_left", "Ctrl+L", this,
      SLOT(slotEditAlignLeft()), actionCollection(), "edit_align_left");
   actionEditAlignRight = new KToggleAction(i18n("Align &right"), "text_right", "Ctrl+R",
      this, SLOT(slotEditAlignRight()), actionCollection(), "edit_align_right");
   actionEditAlignJustify = new KToggleAction(i18n("&Justify"), "text_block", "Ctrl+J",
      this, SLOT(slotEditAlignJustify()), actionCollection(), "edit_align_justify");
   actionEditAlignCenter = new KToggleAction(i18n("&Center"), "text_center", "Ctrl+E", this,
      SLOT(slotEditAlignCenter()), actionCollection(), "edit_align_center");
   actionEditDate = new KAction(i18n("Insert &date"), "", "F5", this,
      SLOT(slotEditInsertDate()), actionCollection(), "edit_insert_date");
   actionEditFile = new KAction(i18n("Insert &file..."), "", "", this,
      SLOT(slotEditInsertFile()), actionCollection(), "edit_insert_file");
   actionEditSwitch = new KAction(i18n("Switch area"), "", "Ctrl+M", this,
      SLOT(slotEditSwitch()), actionCollection(), "edit_switch");
   actionRawTextMode = new KToggleAction(i18n("&Raw Text Mode"),
       "source", "F10", this, SLOT(slotRawTextMode()),
       actionCollection(), "edit_raw_text");

   actionNoteAdd = new KAction(i18n("&Add"), "", "Alt+Insert", this,
      SLOT(slotNoteAdd()), actionCollection(), "note_add");
   actionNoteAddChild = new KAction(i18n("Add &subnote"), "", "Alt+Shift+Insert", this,
      SLOT(slotNoteAddChild()), actionCollection(), "note_add_child");
   actionNoteRemove = new KAction(i18n("&Delete"), "", "Alt+Delete", this,
      SLOT(slotNoteRemove()), actionCollection(), "note_delete");
   actionNoteRename = new KAction(i18n("&Rename"), "", "F2", this,
      SLOT(slotNoteRename()), actionCollection(), "note_rename");
   actionNoteExpandAll = new KAction(i18n("E&xpand all"), "", "", this,
      SLOT(slotNoteExpandAll()), actionCollection(), "note_expand_all");
   actionNoteExpand = new KAction(i18n("&Expand current"), "", "Alt++", this,
      SLOT(slotNoteExpand()), actionCollection(), "note_expand");
   actionNoteCollapseAll = new KAction(i18n("Co&llapse all"), "", "", this,
      SLOT(slotNoteCollapseAll()), actionCollection(), "note_collapse_all");
   actionNoteCollapse = new KAction(i18n("&Collapse current"), "", "Alt+-", this,
      SLOT(slotNoteCollapse()), actionCollection(), "note_collapse");
   actionNoteMoveUp = new KAction(i18n("Move up"), "", "Alt+Shift+Up", this,
      SLOT(slotNoteMoveUp()), actionCollection(), "note_move_up");
   actionNoteMoveDown = new KAction(i18n("Move down"), "", "Alt+Shift+Down", this,
      SLOT(slotNoteMoveDown()), actionCollection(), "note_move_down");
   actionNoteMoveRight = new KAction(i18n("Move level up"), "", "Alt+Shift+Right", this,
      SLOT(slotNoteMoveRight()), actionCollection(), "note_move_right");
   actionNoteMoveLeft = new KAction(i18n("Move level down"), "", "Alt+Shift+Left", this,
      SLOT(slotNoteMoveLeft()), actionCollection(), "note_move_left");
   actionNoteMoveBegin = new KAction(i18n("Move at the beginning"), "", "Alt+Shift+Home", this,
      SLOT(slotNoteMoveBegin()), actionCollection(), "note_move_begin");
   actionNoteMoveEnd = new KAction(i18n("Move at the end"), "", "Alt+Shift+End", this,
      SLOT(slotNoteMoveEnd()), actionCollection(), "note_move_end");

   actionNoteSort = new KAction(i18n("Sort"), "", "", this,
      SLOT(slotNoteSort()), actionCollection(), "note_sort");

   actionOptions = KStdAction::preferences(this, SLOT(slotOptions()),
      actionCollection(), "options_configure");
   actionOptionsKeys = KStdAction::keyBindings(this, SLOT(slotOptionsKeys()),
      actionCollection(), "options_keys");
   actionOverwriteMode = new KAction(i18n("Overwrite Mode"), "", "Insert", this,
      SLOT(slotOverwriteModeChange()), actionCollection(), "overwrite_mode_change");
   KAction* actionHelpTip = new KAction(i18n("&Tip of the day"), "idea", "", this,
      SLOT(slotHelpTip()), actionCollection(), "help_tip");

   actionNoteUp = new KAction(i18n("Go to previous note"), "", "Alt+Up", this,
      SLOT(slotNoteUp()), actionCollection(), "note_previous");
   actionNoteDown = new KAction(i18n("Go to next note"), "", "Alt+Down", this,
      SLOT(slotNoteDown()), actionCollection(), "note_next");
   actionNoteBegin = new KAction(i18n("Go to first note"), "", "Alt+Home", this,
      SLOT(slotNoteBegin()), actionCollection(), "note_begin");
   actionNoteEnd = new KAction(i18n("Go to last note"), "", "Alt+End", this,
      SLOT(slotNoteEnd()), actionCollection(), "note_end");
   actionNoteRight = new KAction(i18n("Go to first subnote"), "", "Alt+Right", this,
      SLOT(slotNoteRight()), actionCollection(), "note_right");
   actionNoteLeft = new KAction(i18n("Go to parent note"), "", "Alt+Left", this,
      SLOT(slotNoteLeft()), actionCollection(), "note_left");
   
      
   KPopupMenu* file = new KPopupMenu(this);
   menuBar()->insertItem(i18n("&File"), file );
   actionNew->plug(file);
   actionNew->plug(toolBar());
   actionOpen->plug(file);
   actionOpen->plug(toolBar());
   actionRecent->plug(file);
   actionSave->plug(file);
   actionSave->plug(toolBar());
   actionSaveAs->plug(file);
   actionFileExport->plug(file);
   actionPrint->plug(file);
   actionFileInfo->plug(file);
   file->insertSeparator();
   actionQuit->plug(file);
   toolBar()->insertSeparator();
   
   KPopupMenu* edit = new KPopupMenu(this);
   menuBar()->insertItem(i18n("&Edit"), edit);
   actionUndo->plug(edit);
   actionUndo->plug(toolBar());
   actionRedo->plug(edit);
   actionRedo->plug(toolBar());
   edit->insertSeparator();
   toolBar()->insertSeparator();
   actionCut->plug(edit);
   actionCut->plug(toolBar());
   actionCopy->plug(edit);
   actionCopy->plug(toolBar());
   actionPaste->plug(edit);
   actionPaste->plug(toolBar());
   edit->insertSeparator();
   toolBar()->insertSeparator();
   KPopupMenu* editfont = new KPopupMenu(this);
   edit->insertItem(i18n("F&ont"), editfont);
   actionBold->plug(editfont);
   actionBold->plug(toolBar());
   actionItalic->plug(editfont);
   actionItalic->plug(toolBar());
   actionUnderline->plug(editfont);
   actionUnderline->plug(toolBar());
   actionColor->plug(editfont);
   actionColor->plug(toolBar());
   editfont->insertSeparator();
   actionSuperscript->plug(editfont);
   actionSubscript->plug(editfont);   
   actionVNormal->plug(editfont);
   KPopupMenu* editpar = new KPopupMenu(this);
   edit->insertItem(i18n("&Paragraph"), editpar);
   actionEditAlignLeft->plug(editpar);
   actionEditAlignLeft->plug(toolBar());
   actionEditAlignRight->plug(editpar);
   actionEditAlignRight->plug(toolBar());
   actionEditAlignCenter->plug(editpar);   
   actionEditAlignCenter->plug(toolBar());
   actionEditAlignJustify->plug(editpar);
   actionEditAlignJustify->plug(toolBar());
   edit->insertSeparator();
   toolBar()->insertSeparator();

   actionGoTo->plug(edit);
   actionGoTo->plug(toolBar());
   actionFind->plug(edit);
   actionFind->plug(toolBar());
   actionFindNext->plug(edit);
   actionFindNext->plug(toolBar());
   actionEditDate->plug(edit);
   actionEditFile->plug(edit);
   actionRawTextMode->plug(edit);
   
   NotesMenu = new KPopupMenu(this);
   menuBar()->insertItem(i18n("&Notes"), NotesMenu);
   actionNoteAdd->plug(NotesMenu);
   actionNoteAddChild->plug(NotesMenu);
   actionNoteRemove->plug(NotesMenu);
   actionNoteRename->plug(NotesMenu);

   KPopupMenu* notesPosition = new KPopupMenu(this);
   NotesMenu->insertItem(i18n("&Position"), notesPosition);
   actionNoteMoveUp->plug(notesPosition);
   actionNoteMoveDown->plug(notesPosition);
   actionNoteMoveBegin->plug(notesPosition);
   actionNoteMoveEnd->plug(notesPosition);
   actionNoteMoveLeft->plug(notesPosition);
   actionNoteMoveRight->plug(notesPosition);
   
   KPopupMenu* notesTree = new KPopupMenu(this);
   NotesMenu->insertItem(i18n("&Tree"), notesTree);
   actionNoteExpand->plug(notesTree);
   actionNoteExpandAll->plug(notesTree);
   actionNoteCollapse->plug(notesTree);
   actionNoteCollapseAll->plug(notesTree);

   actionNoteSort->plug(NotesMenu);
                   
   KPopupMenu* options = new KPopupMenu(this);
   menuBar()->insertItem(i18n("&Options"), options);
   actionOptions->plug(options);
   actionOptionsKeys->plug(options);
   
   KPopupMenu* help = helpMenu();
   menuBar()->insertItem( i18n("&Help"), help );
   actionHelpTip->plug(help);

   /* insert/overwrite mode */
   statusBar()->insertFixedItem(("WWW"), StatusOvr, true);
   statusBar()->changeItem(i18n("INS"), StatusOvr);
   connect(statusBar(), SIGNAL(pressed(int)), this,
      SLOT(slotOverwriteModeChange(int)));

   /* system tray */
   systray = new KnowitTray(this, "Systray");
   systray->setPixmap(Notes.Pixmaps[1]);
   systray->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);

   /* autosave */
   AutosaveTimer = new QTimer(this);
   connect(AutosaveTimer, SIGNAL(timeout()), this, SLOT(slotFileSave()));

   /* configuration */
   config = kapp->config();
   readOptions();

   if (runMinimized)
      showMinimized();
   else
       show();

   bool opened = false;
   KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
   for(int i = 0; i < args->count(); i++)
       if (open(args->url(i))) {
          opened = true; break;
          }
   args->clear();
          
   if (!opened && Options.reopen && filename.path() != "")
      open(KURL(filename));

   slotActionUpdate();

   KTipDialog::showTip(this);
}

Knowit::~Knowit()
{
}

bool Knowit::open(const KURL& fname)
{
   QFile file(fname.path());
   if (!file.open(IO_ReadOnly)) {
      KMessageBox::error(0, i18n("<qt>Cannot open file<br><b>%1</b></qt>")
         .arg(fname.url()));
      if (filename == fname)
         filename = "";
      return false;
      }

   prevItem = 0;
   uint prevdepth = 0;
   reset();
   QTextStream ts(&file);
   ts.setEncoding(QTextStream::UnicodeUTF8);
   QString s, text;
   QListViewItem* current = 0;
   QListViewItem* active = 0;
   Notes.autoUpdate = false;
   while (true) {
      s = ts.readLine();
      if (s.isNull()) {
         if (current)
            Notes.addNote(current, text);
         break;
         }
      if (s.left(9) == "\\NewEntry" || s.left(13) == "\\CurrentEntry") {
         if (current)
            Notes.addNote(current, text);
         QString name;
         uint depth;
         QTextIStream ls(&s);
         ls >> name >> depth;
         name = ls.readLine();
         name = name.stripWhiteSpace();
         if (depth == prevdepth) /* same level */
            if (depth)
               current = new KListViewItem(current->parent(), current, name);
            else
               current = new KListViewItem(Items, current, name);
         else if (depth > prevdepth)   /* subnode */
            current = new KListViewItem(current, name);
         else {   /* higher node */
            for(uint i=depth; i<prevdepth && current; i++)
               current = current->parent();
            if (current && current->parent())
               current = new KListViewItem(current->parent(), current, name);
            else if (current)
               current = new KListViewItem(Items, current, name);
            else
               current = new KListViewItem(Items, current, name);
         }
         if (s.left(13) == "\\CurrentEntry")
            active = current;
         text = "";
         prevdepth = depth;
      }
      else {
         if (text.length())
            text += '\n';
         text += s;
      }
   }
   Notes.autoUpdate = true;
   Notes.updateNotesView();
   filename = fname;
   setCaption(fname.fileName());
   Notes.modified = false;
   if (active)
      showItem(active);
   else
      Edit->clear();
   Edit->setModified(false);
   actionRecent->addURL(filename);
   slotStatusMsg(i18n("File %1 opened.").arg(filename.fileName()));
   slotActionUpdate();
   return true;
}

bool Knowit::save(const KURL& fname)
{
  QFile file(fname.path());
  if (!file.open( IO_WriteOnly)) {
     KMessageBox::error(0, i18n("<qt>Cannot save file<br><b>%1</b></qt>")
         .arg(fname.url()));
     return false;
     }
  if (prevItem && Edit->isModified())
     Notes.changeNote(prevItem, Edit->length() ? Edit->text() : QString());
  QTextStream ts(&file);
  ts.setEncoding(QTextStream::UnicodeUTF8);
  QString header = "\\NewEntry %1 %2\n";
  QString activeheader = "\\CurrentEntry %1 %2\n";
  QListViewItemIterator it(Items);
  while (it.current()) {
     if (Items->currentItem() == it.current())
        ts << activeheader.arg(it.current()->depth()).arg(it.current()->text(0));
     else
        ts << header.arg(it.current()->depth()).arg(it.current()->text(0));
     ts << Notes.text(it.current()) << '\n';
     it++;
  }
  filename = fname;
  setCaption(fname.fileName());
  Notes.modified = false;
  Edit->setModified(false);
  actionRecent->addURL(filename);
  slotStatusMsg(i18n("File %1 saved.").arg(filename.fileName()));
  if (Options.autosave)
     AutosaveTimer->start(Options.autosave * 60 * 1000, true);
  return true;
}

void Knowit::reset()
{
   prevItem = 0;
   Notes.clearNotes();
   Items->clear();
   Edit->clear();
   filename = "";
   setCaption(Untitled);
   Notes.modified = false;
   slotActionUpdate();
}



bool Knowit::queryClose()
{
   if (!modified()) return true;
   else switch (KMessageBox::questionYesNoCancel(0,
       i18n("<qt>File <b>%1</b>\nwas modified. Do you want to save it?</qt>")
          .arg(filename.isEmpty() ? Untitled : filename.fileName()))) {
      case KMessageBox::Yes:
         slotFileSave();
         return !modified();
      case KMessageBox::No:
         return true;
      case KMessageBox::Cancel:
         return false;
      }
   return false;
}

bool Knowit::queryExit()
{
   saveOptions();
   return false;
}


void Knowit::find(QListViewItem* start)
{
   if (soughtText.isEmpty() || !Items->firstChild()) return;
   if (start && lastSought == start &&
      Edit->find(soughtText, FindDlg.case_sensitive(), false))
      return;
   if (!start) start = Items->firstChild();
   QListViewItemIterator it(start);
   while (it.current()) {
      QString s = Notes.text(it.current());
      if (lastSought != it.current() &&
         s.contains(soughtText, FindDlg.case_sensitive())) {
         showItem(it.current());
         Edit->find(soughtText, FindDlg.case_sensitive(), false);
         lastSought = it.current();
         return;
      }
      it++;
   }
   KMessageBox::information(0, i18n("<qt>Sought text:<br><b>%1</b><br>not found.</qt>").arg(soughtText));
}


void Knowit::saveOptions()
{
  config->setGroup("MainWindow");
  config->writeEntry("Widgets", Layout->sizes());
  Options.save(config);
  config->setGroup("General");
  config->writeEntry("Minimized", isMinimized() || !isVisible());
  config->writeEntry("LastFile", filename.path());
  actionCollection()->writeShortcutSettings("Keys", config);
  config->deleteGroup("Recent Files");
  actionRecent->saveEntries(config, "Recent Files");
}

void Knowit::readOptions()
{
   config->setGroup("MainWindow");
   QValueList<int> sizes = config->readIntListEntry("Widgets");
   if (sizes.count() == 2)
      Layout->setSizes(sizes);
   Options.read(config);
   config->setGroup("General");
   runMinimized = config->readBoolEntry("Minimized", false);
   actionCollection()->readShortcutSettings("Keys", config);
   actionRecent->loadEntries(config,"Recent Files");
   QString oldfile = config->readEntry("LastFile");
   if (!oldfile.isEmpty())
      filename = KURL(oldfile);
   applyOptions(Options);
}

void Knowit::showItem(QListViewItem* item)
{
   if (item) {
      if (Options.autoCollapse)
         slotNoteCollapseAll();
      Items->setCurrentItem(item);
      Items->setSelected(item, true);
      Items->ensureItemVisible(item);
   }
}

bool Knowit::modified()
{
   return Notes.modified || (Items->currentItem() && Edit->isModified());
}

void Knowit::closeEvent(QCloseEvent* e)                          
{
   hide();
   KMainWindow::closeEvent(e);
}

void Knowit::applyOptions(const KnowitOptions& O, bool store)
{
   if (O.docked)
      systray->show();
   else
      systray->hide();
   if (O.toolbar)
      toolBar()->show();
   else
      toolBar()->hide();
   Edit->setWordWrap(O.wordwrap ? QTextEdit::WidgetWidth : QTextEdit::NoWrap);
   if (!O.autosave)
       AutosaveTimer->stop();
   else
       AutosaveTimer->start(O.autosave * 60 * 1000, true);
   Layout->setOrientation(O.horizontalSplit ? QSplitter::Horizontal :
         QSplitter::Vertical);
   Items->setAlternateBackground(O.alternateTree ? QColor(O.alternateColor) : QColor());
   Items->repaintContents();
   showItem(Items->currentItem());
   if (store) Options = O;
}








/*** Slots ***/
void Knowit::slotStatusMsg(const QString &text)
{
   statusBar()->clear();
   statusBar()->changeItem(text, 0);
}


void Knowit::slotContextMenu(KListView*, QListViewItem*, const QPoint& p)
{
   NotesMenu->exec(p);
}





/******* Slots: Notes *******/
void Knowit::slotNoteAdd()
{
   bool Ok;
   KListViewItem* elt;
   QString text = KLineEditDlg::getText(i18n("Add note"),
      i18n("Note title:"), Options.defaultName, &Ok);
   if (!Ok) return;
   if (Items->currentItem())
      if (Items->currentItem()->parent())
         elt = new KListViewItem(Items->currentItem()->parent(),
            Items->currentItem(), text);
      else
         elt = new KListViewItem(Items, Items->currentItem(), text);
   else
      elt = new KListViewItem(Items, text);
   Notes.addNote(elt);
   showItem(elt);
   if (Notes.count() == 1)
      slotActionUpdate();
}

void Knowit::slotNoteAddChild()
{
   if (!Items->currentItem()) {
      KMessageBox::error(0, i18n("You have to choose note before adding subnote."));
      return;
   }
   bool Ok;         
   QString text = KLineEditDlg::getText(i18n("Add subnote"),
      i18n("Note title:"), i18n("New subnote"), &Ok);
   if (!Ok) return;
   QListViewItem* lastChild = Items->currentItem()->firstChild();
   while (lastChild && lastChild->nextSibling())
      lastChild = lastChild->nextSibling();

   KListViewItem* elt;
   if (lastChild)
       elt = new KListViewItem(Items->currentItem(), lastChild, text);
   else
      elt = new KListViewItem(Items->currentItem(), text);
   Notes.addNote(elt);
   showItem(elt);
}

void Knowit::slotNoteRemove()
{
   QListViewItem* elt = Items->currentItem();
   if (!elt) return;
   QString msg = elt->childCount() ?
      i18n("<qt>Are you sure you want to delete note<br><b>%1</b><br> "
         "and its subnotes?</qt>") :
      i18n("<qt>Are you sure you want to delete note<br><b>%1</b>?</qt>");
   if (KMessageBox::questionYesNo(0, msg.arg(elt->text(0))) == KMessageBox::Yes) {
      QListViewItem* parent = elt->parent();
      Notes.removeNote(elt);
      if (!Notes.count())
         slotActionUpdate();
      if (!parent)
         parent = Items->firstChild();
      prevItem = 0;
      showItem(parent);         
      }
}

void Knowit::slotNoteRename()
{
    if (Items->currentItem()) {
       Items->currentItem()->setRenameEnabled(0, true);
       Items->currentItem()->startRename(0);
       }
}    

void Knowit::slotNoteCollapseAll()
{
   QListViewItem* it = Items->firstChild();
   while(it) {
      it->setOpen(false);
      it = it->nextSibling();
      }
}

void Knowit::slotNoteCollapse()
{
   if (Items->currentItem()) Items->currentItem()->setOpen(false);
}


void Knowit::slotNoteExpandAll()
{
   QListViewItem* it = Items->firstChild();
   while(it) {
      it->setOpen(true);
      it = it->nextSibling();
      }
}

void Knowit::slotNoteExpand()
{
   if (Items->currentItem()) Items->currentItem()->setOpen(true);
}








void Knowit::slotItemChanged(QListViewItem* i)
{
   if (prevItem && Edit->isModified())
      Notes.changeNote(prevItem, Edit->length() ? Edit->text() : QString());
   prevItem = i;
   QString s;
   if (!prevItem || (s = Notes.text(prevItem)).isEmpty())
      Edit->setText("");
   else
      Edit->setText(s);
   Edit->setModified(false);
}

void Knowit::slotNoteMoveUp()
{
   QListViewItem* elt = Items->currentItem();
   if (elt && elt->itemAbove()) {
      QListViewItem* above = elt->parent() ? elt->parent()->firstChild()
         : Items->firstChild();
      if (above == elt)
         return;
      if (!above->nextSibling() || above->nextSibling() == elt)
         above = 0;
      else while (above->nextSibling()->nextSibling() != elt)
         above = above->nextSibling();
      Items->moveItem(elt, elt->parent(), above);
      showItem(elt);
      Notes.modified = true;
      }
}

void Knowit::slotNoteMoveDown()
{
   QListViewItem* elt = Items->currentItem();
   if (elt && elt->nextSibling()) {
      Items->moveItem(elt, elt->parent(), elt->nextSibling());
      showItem(elt);
      Notes.modified = true;
      }
}

void Knowit::slotNoteMoveBegin()
{
   QListViewItem* elt = Items->currentItem();
   if (elt) {
      Items->moveItem(elt, elt->parent(), 0);
      showItem(elt);
      Notes.modified = true;
      }
}

void Knowit::slotNoteMoveEnd()
{
   QListViewItem* elt = Items->currentItem();
   if (!elt)
      return;
   QListViewItem* last = elt;
   while (last->nextSibling())
      last = last->nextSibling();
   if (elt && elt != last) {
      Items->moveItem(elt, elt->parent(), last);
      showItem(elt);
      Notes.modified = true;
      }
}

void Knowit::slotNoteMoveLeft()
{
   QListViewItem* elt = Items->currentItem();
   if (elt && elt->parent()) {
      Items->moveItem(elt, elt->parent()->parent(), elt->parent());
      showItem(elt);
      Notes.modified = true;
      }
}


void Knowit::slotNoteMoveRight()
{
   QListViewItem* elt = Items->currentItem();
   if (elt && elt->itemAbove()) {
      if (elt->depth() == elt->itemAbove()->depth())
         Items->moveItem(elt, elt->itemAbove(), 0);
      else 
         Items->moveItem(elt, elt->itemAbove()->parent(), elt->itemAbove());
      showItem(elt);
      Notes.modified = true;
   }
}


void Knowit::slotNoteSort()
{
   if (Items->currentItem()) {
      Items->currentItem()->sortChildItems(0, true);
      Items->repaintContents();
   }
}

void Knowit::slotNoteUp()
{
   if (Items->currentItem())
      showItem(Items->currentItem()->itemAbove());
}

void Knowit::slotNoteDown()
{
   if (Items->currentItem())
      showItem(Items->currentItem()->itemBelow());
}

void Knowit::slotNoteBegin()
{
   showItem(Items->firstChild());
}

void Knowit::slotNoteEnd()
{
   QListViewItem* elt = Items->firstChild();
   if (elt)
      while (elt->itemBelow())
         elt = elt->itemBelow();
   showItem(elt);
}

void Knowit::slotNoteLeft()
{
   if (Items->currentItem())
      showItem(Items->currentItem()->parent());
}

void Knowit::slotNoteRight()
{
   if (Items->currentItem())
      showItem(Items->currentItem()->firstChild());
}



void Knowit::slotFileNew()
{
   if (queryClose()) reset();
}

void Knowit::slotFileOpen()
{
   if (!queryClose()) return;
   KURL url=KFileDialog::getOpenURL(QString::null,
      i18n("*.kno|KnowIt files (*.kno)\n*|All files"), this, i18n("Open File..."));
   if (!url.isEmpty())
      open(url);
}

void Knowit::slotFileSave()
{
  if (filename.isEmpty())
     slotFileSaveAs();
  else {
     save(filename);
     }
}

void Knowit::slotFileRecent(const KURL& url)
{
   if (queryClose())
      if (!open(url))
         actionRecent->removeURL(url);
}


void Knowit::slotFileSaveAs()
{
   KURL url=KFileDialog::getSaveURL(QDir::currentDirPath(),
        i18n("*.kno|KnowIt files (*.kno)\n*|All files"), this, i18n("Save as..."));
   if (!url.isEmpty()) {
      if (!url.fileName().contains('.'))
        url = KURL(url.path() + ".kno");
      QFileInfo fileinfo(url.path());
      if (fileinfo.exists() && KMessageBox::questionYesNo(0,
         i18n("<qt>File<br><b>%1</b><br>already exists. Overwrite it?</qt>")
            .arg(url.path())) == KMessageBox::No)
         return;
      save(url);
  }
}


void Knowit::slotFileInfo()
{
   int count = 0;
   QListViewItemIterator it(Items);
   while (it.current()) {
      count++;
     it++;
   }
   KMessageBox::information(0, i18n("<qt><h1>%1</h1>"
      "Document path: %2<br>"
      "Number of notes: %3</qt>").arg(filename.fileName())
         .arg(filename.path()).arg(count));
}


void Knowit::slotFilePrint()
{
   KPrinter printer;
   printer.setFullPage(true);
   if (Edit->text().stripWhiteSpace().isEmpty() || !printer.setup(this))
      return;
      
   QSimpleRichText richText(Edit->text(), Edit->font());
   QPainter p(&printer);
   QPaintDeviceMetrics metrics(p.device());
   int dpix = metrics.logicalDpiX();
   int dpiy = metrics.logicalDpiY();
   const int margin = 72; // pt
   QRect body(margin*dpix / 72, margin*dpiy / 72,
      metrics.width()-  margin*dpix /72*2,
      metrics.height()-margin*dpiy/72*2);
   richText.setWidth(&p, body.width());
   QRect view(body);
   int page = 1;
   while (true) {
      richText.draw(&p, body.left(), body.top(), view, colorGroup());
         view.moveBy(0, body.height());
      p.translate( 0 , -body.height());
      p.setFont(Edit->font());
      p.drawText( view.right() - p.fontMetrics().width( QString::number(page)),
         view.bottom() + p.fontMetrics().ascent() + 5, QString::number(page));
      if (view.top() >= body.top() + richText.height())
         break;
      printer.newPage();
      page++;
   } 
}




void Knowit::slotFileQuit()
{
   if (queryClose())
      kapp->quit();
}

void Knowit::slotFileExport()
{
  if (!Items->firstChild()) return;         /* nothing to export */
  if (ChooserDlg.exec() != QDialog::Accepted) return; /* cancelled */

  /* Deduce filename from Knowit document name */
  QString htmlfile;
  if (filename.isValid()) {
     htmlfile = filename.path();
     if (htmlfile.right(4) == ".kno")
        htmlfile = htmlfile.left(htmlfile.length() - 4) + ".html";
     else htmlfile = "";
  }
  if (htmlfile.isEmpty())
     htmlfile == QDir::currentDirPath();

  /* Confirm/choose filename */  
  KURL url=KFileDialog::getSaveURL(htmlfile,
        i18n("*.html|HTML files (*.html)\n*|All files"), this, i18n("Export to HTML"));
  if (!url.isEmpty()) {
     if (!url.fileName().contains('.'))
        url = KURL(url.path() + ".html");
     QFileInfo fileinfo(url.path());
     if (fileinfo.exists() && KMessageBox::questionYesNo(0,
         i18n("<qt>File<br><b>%1</b><br>already exists. Overwrite it?</qt>")
            .arg(url.path())) == KMessageBox::No)
        return;
     QListViewItem* start = (ChooserDlg.getChoice() == KnowitChooser::All) ?
         Items->firstChild() : Items->currentItem();
     Notes.find(start)->saveHTML(url, ChooserDlg.getChoice());
  }
}

void Knowit::slotEditBold()
{
   Edit->setBold(actionBold->isChecked());
}

void Knowit::slotEditItalic()
{
   Edit->setItalic(actionItalic->isChecked());
}

void Knowit::slotEditUnderline()
{
   Edit->setUnderline(actionUnderline->isChecked());
}

void Knowit::slotEditColor()
{
   QColor color = Edit->color();
   if (KColorDialog::getColor(color) == QDialog::Accepted)
      Edit->setColor(color);
}

void Knowit::slotEditSuperscript()
{
   Edit->setVerticalAlignment(QTextEdit::AlignSuperScript);
}

void Knowit::slotEditSubscript()
{
   Edit->setVerticalAlignment(QTextEdit::AlignSubScript);
}

void Knowit::slotEditNormal()
{
   Edit->setVerticalAlignment(QTextEdit::AlignNormal);
}

void Knowit::slotEditAlignLeft()
{
   Edit->setAlignment(Qt::AlignLeft);
}

void Knowit::slotEditAlignRight()
{
   Edit->setAlignment(Qt::AlignRight);
}

void Knowit::slotEditAlignJustify()
{
   Edit->setAlignment(Qt::AlignJustify);
}

void Knowit::slotEditAlignCenter()
{
   Edit->setAlignment(Qt::AlignCenter);
}

void Knowit::slotEditFind()
{
   if (soughtText != FindDlg.getText()) {
      soughtText = FindDlg.getText();
      lastSought = 0;
      find();
      }
   else 
      find(Items->currentItem());
}

void Knowit::slotEditFindNext()
{
   if (soughtText.isEmpty()) FindDlg.exec();
   find(Items->currentItem());
}

void Knowit::slotEditGoTo()
{
   bool Ok;
   QString page = KLineEditDlg::getText(i18n("Go to"),
      i18n("Go to page with given title"), "", &Ok);
   if (Ok && !page.isEmpty()) {
      QListViewItem* item = Items->findItem(page, 0);
      if (!item) item = Items->findItem(page, 0, Qt::ExactMatch);
      if (!item) item = Items->findItem(page, 0, Qt::Contains);
      showItem(item);
   }   
}

void Knowit::slotEditInsertDate()
{
   if (Items->currentItem()) {
      bool ib = actionBold->isChecked(), ii=actionItalic->isChecked(), iu=actionUnderline->isChecked();
      QColor color = Edit->color();
      Edit->setItalic(Options.insertDateItalic);
      Edit->setBold(Options.insertDateBold);
      Edit->setUnderline(Options.insertDateUnderline);
      Edit->setColor(QColor(Options.insertDateColor));
      Edit->insert(QString("%1")
         .arg(QDateTime::currentDateTime().toString(Options.insertDateFormat)));
      Edit->setItalic(ii);
      Edit->setBold(ib);
      Edit->setUnderline(iu);
      Edit->setColor(color);
      
   }
}

void Knowit::slotEditInsertFile()
{
   if (!Items->currentItem())
      return;
   KURL url=KFileDialog::getOpenURL(QString::null,
      i18n("*|All files"), this, i18n("Insert file..."));
   if (url.isEmpty())
      return;
   QFile file(url.path());
   if (!file.open(IO_ReadOnly)) {
      KMessageBox::error(0, i18n("<qt>Cannot open file<br><b>%1</b></qt>")
         .arg(url.url()));
      return;
      }
   QTextStream ts(&file);
   ts.setEncoding(QTextStream::UnicodeUTF8);
   QString s, text;
   while (true) {
      s = ts.readLine();
      if (s.isNull())
         break;
      if (s.isEmpty())
         text += '\n';
      else
         text += s;
      }
   Edit->insert(text);
}


void Knowit::slotEditSwitch()
{
   if (Items->hasFocus())
      Edit->setFocus();
   else Items->setFocus();
}

void Knowit::slotEditChanged()
{
   actionBold->setChecked(Edit->bold());
   actionItalic->setChecked(Edit->italic());
   actionUnderline->setChecked(Edit->underline());
   actionEditAlignLeft->setChecked(Edit->alignment() == Qt::AlignLeft);
   actionEditAlignRight->setChecked(Edit->alignment() == Qt::AlignRight);
   actionEditAlignCenter->setChecked(Edit->alignment() == Qt::AlignCenter);
   actionEditAlignJustify->setChecked(Edit->alignment() == Qt::AlignJustify);
}

void Knowit::slotEditCursorChanged(int, int)
{
   slotEditChanged();
}


void Knowit::slotOptions()
{
   PrefDlg.setOptions(Options);
   if (PrefDlg.exec() == QDialog::Accepted)
      applyOptions(PrefDlg.getOptions());
}

void Knowit::slotOptionsKeys()
{
   KKeyDialog K(false);
   K.configure(actionCollection());
}


void Knowit::slotUndock()
{
   Options.docked = false;
   applyOptions(Options);
}

void Knowit::slotHelpTip()
{
   KTipDialog::showTip(this, 0, true);
}


void Knowit::slotActionUpdate()
{
   /* todo: optimization - invoke only when neccessary */
   bool editable = Items->childCount();
   Edit->setReadOnly(!editable);
   actionBold->setEnabled(editable);
   actionItalic->setEnabled(editable);
   actionUnderline->setEnabled(editable);
   actionColor->setEnabled(editable);
   actionEditDate->setEnabled(editable);
   actionEditFile->setEnabled(editable);   
   actionEditAlignLeft->setEnabled(editable);
   actionEditAlignRight->setEnabled(editable);
   actionEditAlignCenter->setEnabled(editable);
   actionEditAlignJustify->setEnabled(editable);
   actionSuperscript->setEnabled(editable);
   actionSubscript->setEnabled(editable);
   actionVNormal->setEnabled(editable);   
   actionRawTextMode->setEnabled(editable);
   actionFind->setEnabled(editable);
   actionFindNext->setEnabled(editable);
   actionGoTo->setEnabled(editable);
   actionNoteAddChild->setEnabled(editable);
   actionNoteRemove->setEnabled(editable);
   actionNoteRename->setEnabled(editable);
   actionNoteCollapseAll->setEnabled(editable);
   actionNoteCollapse->setEnabled(editable);
   actionNoteExpandAll->setEnabled(editable);
   actionNoteExpand->setEnabled(editable);
   actionNoteMoveUp->setEnabled(editable);
   actionNoteMoveDown->setEnabled(editable);
   actionNoteMoveBegin->setEnabled(editable);
   actionNoteMoveEnd->setEnabled(editable);
   actionNoteMoveLeft->setEnabled(editable);
   actionNoteMoveRight->setEnabled(editable);
}

void Knowit::slotOverwriteModeChange(int i)
{
   if (i == StatusOvr)
      slotOverwriteModeChange();
}

void Knowit::slotOverwriteModeChange()
{
   Edit->setOverwriteMode(!Edit->isOverwriteMode());
   statusBar()->changeItem(Edit->isOverwriteMode() ? i18n("OVR") :
      i18n("INS"), StatusOvr);
}

void Knowit::slotRawTextMode()
{
   if (prevItem && Edit->isModified())
      Notes.changeNote(prevItem, Edit->length() ? Edit->text() : QString());
   Edit->setTextFormat(actionRawTextMode->isChecked() ?
      Qt::PlainText : Qt::RichText);
   Edit->setText(Notes.text(Items->selectedItem()));
}

